/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package eu.agrosense.client.geotiff;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import java.awt.Image;
import java.beans.PropertyChangeListener;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.geoviewer.Geographical;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.processing.CoverageProcessor;
import org.geotools.geometry.jts.JTS;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.geometry.BoundingBox;
import org.opengis.geometry.Envelope;
import org.opengis.parameter.ParameterValueGroup;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.TransformException;
import org.openide.util.Exceptions;

/**
 *
 * @author johan
 */
public abstract class AbstractGeoTiffGeographical<T> implements Geographical<T> {
    private static final Logger LOGGER = Logger.getLogger("eu.agrosense.client.geotiff.AbstractGeoTiffGeographical");
    
    protected final GridCoverage2D coverage;
    protected final Geometry boundingBox;

    public AbstractGeoTiffGeographical(GridCoverage2D coverage) {
        this.coverage = coverage;
        this.boundingBox = getWGS84BoundingBox(coverage);
    }

    @Override
    public Point getCentroid() {
        return boundingBox != null ? boundingBox.getCentroid() : null;
    }

    @Override
    public Geometry getBoundingBox() {
        return boundingBox;
    }

    @Override
    public String getTooltipText() {
        return "";
    }
    
    @Override
    public Image getIcon() {
        return null;
    }

    @Override
    public String getIconLabel() {
        return "";
    }

    @Override
    public void addPropertyChangeListener(PropertyChangeListener listener) {
    }

    @Override
    public void removePropertyChangeListener(PropertyChangeListener listener) {
    }


    /**
     * Get the bounding box geometry of a coverage in the WGS84 CRS.
     * 
     * @param coverage
     * @return 
     */
    public static Geometry getWGS84BoundingBox(GridCoverage2D coverage) {
        CoordinateReferenceSystem crs = coverage.getCoordinateReferenceSystem2D();
        LOGGER.log(Level.FINEST, "source CRS: {0}", crs.toString());
        Envelope originalEnvelope = coverage.getEnvelope();
        ReferencedEnvelope referencedEnvelope = new ReferencedEnvelope(originalEnvelope);
        LOGGER.log(Level.FINEST, "source envelope: {0}", referencedEnvelope.toString());

        try {
            ReferencedEnvelope wgs84Envelope = referencedEnvelope.transform(DefaultGeographicCRS.WGS84, true);
            LOGGER.log(Level.FINEST, "wgs84 envelope: {0}", wgs84Envelope.toString());
            return JTS.toGeometry((BoundingBox) wgs84Envelope);
        } catch (FactoryException | TransformException ex) {
            Exceptions.printStackTrace(ex);
            return null;
        }
    }
    
    /**
     * Crop a coverage using a JTS envelope.
     * E.g. using the map window viewport to crop to the visible part of the coverage.
     * 
     * @param coverage
     * @param envelope
     * @return the cropped coverage, or null when there was no intersection, or the original full coverage when cropping wasn't possible.
     */
    public static GridCoverage2D crop(GridCoverage2D coverage, com.vividsolutions.jts.geom.Envelope envelope) {
        if (envelope != null) {
            ReferencedEnvelope coverageEnvelope = new ReferencedEnvelope(coverage.getEnvelope2D());
            ReferencedEnvelope viewportEnvelope = new ReferencedEnvelope(envelope, DefaultGeographicCRS.WGS84);

            try {
                viewportEnvelope = viewportEnvelope.transform(coverage.getCoordinateReferenceSystem2D(), true);
            } catch (TransformException | FactoryException | IllegalArgumentException ex) {
                // IllegalArgumentException: caused by extreme values from transformation when the 
                // source CRS uses a local datum and the map window is showing e.g. another continent.
                LOGGER.log(Level.INFO, "Could not transform viewport to coverage CRS, skip cropping", ex);
                return coverage;
            }

            if (viewportEnvelope.intersects((BoundingBox)coverageEnvelope)) {
                final CoverageProcessor processor = CoverageProcessor.getInstance();
                final ParameterValueGroup params = processor.getOperation("CoverageCrop").getParameters();
                params.parameter("Source").setValue(coverage);
                params.parameter("Envelope").setValue(viewportEnvelope);
                GridCoverage2D cropped = (GridCoverage2D) processor.doOperation(params);
                return cropped;
            } else {
                // nothing to show, coverage is outside viewport
                return null;
            }
        }
        return coverage;        
    }
    
    public static class CroppedGridImpl extends AbstractGeoTiffGeographical<GridCoverage2D> {

        public CroppedGridImpl(GridCoverage2D coverage) {
            super(coverage);
        }

        @Override
        public GridCoverage2D getRenderObject(com.vividsolutions.jts.geom.Envelope envelope) {
            return crop(coverage, envelope);
        }
    }
    
}
