/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package eu.agrosense.client.geotiff;

import com.vividsolutions.jts.geom.Envelope;
import java.awt.Image;
import java.io.IOException;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.geoviewer.BaseLayer;
import nl.cloudfarming.client.geoviewer.RangedLegendPalette;
import nl.cloudfarming.client.geoviewer.SimpleLayer;
import nl.cloudfarming.client.imp.ImportSource;
import nl.cloudfarming.client.util.UOM;
import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.grid.GridCoverage2D;
import org.opengis.coverage.SampleDimensionType;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataNode;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

public class GeoTiffDataObject extends MultiDataObject {
    private static final Logger LOGGER = Logger.getLogger("eu.agrosense.client.geotiff.GeoTiffDataObject");

    public static final String MIME_TYPE = "image/tiff";
    public static final String ICON_BASE = "eu/agrosense/client/geotiff/file_extension_tif.png";

    private final InstanceContent ic;
    private final Lookup intern;
    private final FileObject pf;
    private final GridCoverage2D coverage;
    
    private double[] range;

    public GeoTiffDataObject(FileObject pf, MultiFileLoader loader, List<FileObject> secondary, GridCoverage2D coverage) throws DataObjectExistsException, IOException {
        super(pf, loader);
        registerEditor(MIME_TYPE, true);
        this.ic = new InstanceContent();
        this.intern = new AbstractLookup(this.ic);
        this.pf = pf;
        this.coverage = coverage;
        
        //ImportSource in lookup is now required by ImportAction:
        ic.add(new ImportSource() {
            @Override public Set<FileObject> getFiles() { return files(); }
        });
        
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }   
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    @Override
    public final Lookup getLookup() {
        return new ProxyLookup(super.getLookup(), intern);
    }

    @Override
    protected Node createNodeDelegate() {
        if (associateLookup() >= 1) {
            ic.add(new GeoTiffGeographical());
            DataNode dataNode = new DataNode(this, Children.create(new GeoTiffChildFactory(this), true), 
                    new ProxyLookup(getLookup(), Lookups.fixed(new SimpleLayer(pf.getName()))));
            dataNode.setIconBaseWithExtension(ICON_BASE);
            dataNode.setDisplayName(pf.getName());
            return dataNode;
        }
        return super.createNodeDelegate();
    }
    
    public GridCoverage2D getCoverage() {
        return coverage;
    }
    
    private class GeoTiffGeographical extends AbstractGeoTiffGeographical<Object> {
        
        public GeoTiffGeographical() {
            super(GeoTiffDataObject.this.coverage);
        }

        @Override
        public Object getRenderObject(Envelope envelope) {
            return null;
        }

        @Override
        public Image getIcon() {
            return ImageUtilities.loadImage(ICON_BASE, true);
        }

        @Override
        public String getIconLabel() {
            return pf.getName();
        }
    }
    
    public double[] getRange() {
        if (range == null) {
            range = getRange(coverage);
        }
        return range;
    }
    
    public static RangedLegendPalette<Double> getGridPalette(GeoTiffDataObject dataObject) {
        double[] range = dataObject.getRange();
        return new RangedLegendPalette<>(range[0], range[1], UOM.NOT_SPECIFIED);
    }
    
    protected static double[] getRange(GridCoverage2D coverage) {
        Double min = null, max = null;
        //LOGGER.log(Level.FINEST, "coverage range type: {0}", coverage.getRangeType()); //unsupported...
        LOGGER.log(Level.FINEST, "num dims: {0}", coverage.getNumSampleDimensions());
        
        // using first band, TODO: combine multiple bands for rgb images.
        if (coverage.getNumSampleDimensions() > 0) {
            GridSampleDimension dim = coverage.getSampleDimension(0);
            SampleDimensionType type = dim.getSampleDimensionType();
            LOGGER.log(Level.FINEST, "dim type: {0}", type);
            LOGGER.log(Level.FINEST, "dim range: {0}", dim.getRange());
            double dimMin = dim.getMinimumValue(), dimMax = dim.getMaximumValue();
            LOGGER.log(Level.FINEST, "dim min/max: {0} .. {1}", new Object[] {dimMin, dimMax});
            
            // the only times the ranges were non-null and non-infinite, 
            // it was a single no-data marker value (e.g. 9999, while the actually useful values were in range 0..100)
//            if (dimMin != Double.NEGATIVE_INFINITY) {
//                if (min == null || dimMin < min) { min = dimMin; }
//            };
//            if (dimMax != Double.POSITIVE_INFINITY) {
//                if (max == null || dimMax > max) { max = dimMax; }
//            }

            // fallback:
//            if (min == null || max == null) {
//                if (type == SampleDimensionType.UNSIGNED_8BITS) {
//                    min = 0.0; max = 255.0;
//                } else if (type == SampleDimensionType.SIGNED_16BITS) {
//                    min = -32768.0; max = 32767.0;
//                }//...
//            }
        }
        
        // this actually produces better results in most cases than using the full range supported by the above 16-bit type;
        // we're going to have to scan all values in the tiff however...
        if (min == null || max == null) {  min = 0.0; max = 255.0; }
        
        LOGGER.log(Level.FINEST, "using range: {0} .. {1}", new Object[] {min, max});
        
        return new double[] {min, max};
    }
    
}
