/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package eu.agrosense.client.geotiff;

import java.awt.event.ActionEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.AbstractAction;
import javax.swing.ImageIcon;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.io.imageio.IIOMetadataDumper;
import org.geotools.data.DataSourceException;
import org.geotools.factory.Hints;
import org.geotools.gce.geotiff.GeoTiffReader;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.awt.ActionRegistration;
import org.openide.awt.NotificationDisplayer;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.filesystems.MIMEResolver;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.FileEntry;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiDataObject.Entry;
import org.openide.loaders.MultiFileLoader;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;

/**
 *
 * @author johan
 */
@MIMEResolver.ExtensionRegistration(
    displayName = "#LBL_GeoTiff_LOADER",
mimeType = GeoTiffDataObject.MIME_TYPE,
extension = {"tif", "TIF"})
@DataObject.Registration(
    mimeType = GeoTiffDataObject.MIME_TYPE,
iconBase = GeoTiffDataObject.ICON_BASE,
displayName = "#LBL_GeoTiff_LOADER",
position = 300)
@ActionReferences(value = {
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.OpenAction"), path = "Loaders/image/tiff/Actions", position = 100, separatorAfter = 200),
    @ActionReference(id =
    @ActionID(category = "Edit", id = "org.openide.actions.DeleteAction"), path = "Loaders/image/tiff/Actions", position = 600),
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.RenameAction"), path = "Loaders/image/tiff/Actions", position = 700, separatorAfter = 800),
    @ActionReference(id =
    @ActionID(category = "System", id = "org.openide.actions.ToolsAction"), path = "Loaders/image/tiff/Actions", position = 1300),
    // Geotools viewer action for testing:
//    @ActionReference(id =
//    @ActionID(category = "View", id = "eu.agrosense.client.geotiff.GeotoolsShowAction"), path = "Loaders/image/tiff/Actions", position = 9999),
})
@NbBundle.Messages({
    "LBL_GeoTiff_LOADER=GeoTIFF files",
    "GeoTiffFileLoader load error title=Error loading GeoTIFF file",
    "# {0} - file name",
    "# {1} - error message",
    "GeoTiffFileLoader load error details={0} failed with: {1}",        
})
public class GeoTiffFileLoader extends MultiFileLoader {
    private static final Logger LOGGER = Logger.getLogger("eu.agrosense.client.geotiff.GeoTiffFileLoader");
    private static final ImageIcon ERROR_ICON = ImageUtilities.loadImageIcon("eu/agrosense/client/geotiff/error.png", true);
    
    public static final String PRIMARY_EXT = "tif";
    public static final String[] OPTIONAL_EXT = {"wld", "tfw", "prj"};
    // optional secondary files:
    // - currently supported:
    //   - .wld, .tfw, .prj: plain tiff files can have associated geo metadata in world and projection files 
    //      (so these are not strictly geotiff files, but the combination is supported by the GeoTiffReader)
    // - maybe later:
    //   - .lgo: some geotiffs come with associated listgeo output; does this add anything to the metadata already present within the tiff?
    //   - .ovr: external overviews / pyramids (might use double extension like image.tif.ovr)
    //   - .sld: Styled Layer Descriptor (xml); not specific to geotiff; used in WMS for styling coverages (can contain custom colormap);
    static {
        Arrays.sort(OPTIONAL_EXT);
    }    

    public GeoTiffFileLoader() {
        super(GeoTiffDataObject.class.getName());
    }

    @Override
    protected String actionsContext() {
        StringBuilder builder = new StringBuilder("Loaders/");
        builder.append(GeoTiffDataObject.MIME_TYPE);
        builder.append("/Actions/");
        return builder.toString();
    }
    
    @Override
    protected FileObject findPrimaryFile(FileObject fo) {
        // never recognize folders.
        if (fo.isFolder()) { return null; }
        
        String ext = fo.getExt().toLowerCase();

        // check secondary extensions:
        if (Arrays.binarySearch(OPTIONAL_EXT, ext) >= 0) {
            LOGGER.log(Level.FINEST, "FileObject {0} has secondary extension, creating secondary entry and returning find brother with primary", fo);
            return FileUtil.findBrother(fo, PRIMARY_EXT);
        }
        // check primary extension:
        if (PRIMARY_EXT.equalsIgnoreCase(ext)) { return fo; }
        return null;
    }

    @Override
    protected MultiDataObject createMultiObject(FileObject primaryFile) throws DataObjectExistsException, IOException {
        LOGGER.log(Level.FINEST, "creating multi object for {0}", primaryFile.getNameExt());
        
        GeoTiffReader reader;
        GridCoverage2D coverage;
        try {
            File file = FileUtil.toFile(primaryFile);
            reader = new GeoTiffReader(file, new Hints(Hints.FORCE_LONGITUDE_FIRST_AXIS_ORDER, Boolean.TRUE));
            coverage = reader.read(null);
        } catch (DataSourceException ex) {
            LOGGER.warning(ex.getMessage());
            NotificationDisplayer.getDefault().notify(Bundle.GeoTiffFileLoader_load_error_title(), ERROR_ICON, 
                    Bundle.GeoTiffFileLoader_load_error_details(primaryFile.getNameExt(), ex), null);
            //TODO: move isobus-type InvalidDataObject to utility pkg
            return new MultiDataObject(primaryFile, this);
        }

        LOGGER.log(Level.FINEST, "coverage count: {0}", reader.getGridCoverageCount());
        LOGGER.log(Level.FINEST, "num overviews: {0}", coverage.getNumOverviews()); // requires secondary ovr file
        LOGGER.log(Level.FINEST, "grid geometry: {0}", coverage.getGridGeometry().toString());
        LOGGER.log(Level.FINEST, "original envelope: {0}", reader.getOriginalEnvelope().toString());
        // huge xml-tree containing all tiff tags:
//        IIOMetadataDumper dumper = new IIOMetadataDumper(reader.getMetadata().getRootNode());
//        LOGGER.finest(dumper.getMetadata());
        
        List<FileObject> secondary = new ArrayList<>();
        for(String optional : OPTIONAL_EXT){
            FileObject found = FileUtil.findBrother(primaryFile, optional);
            if(found != null) { secondary.add(found); }
        }
        
        return new GeoTiffDataObject(primaryFile, this, secondary, coverage);
    }

    @Override
    protected Entry createPrimaryEntry(MultiDataObject obj, FileObject primaryFile) {
        assert PRIMARY_EXT.equalsIgnoreCase(primaryFile.getExt());
        Entry entry = new FileEntry(obj, primaryFile);
        LOGGER.log(Level.FINEST, "Creating primary entry for {0}", primaryFile.getNameExt());
        return entry;
    }

    @Override
    protected Entry createSecondaryEntry(MultiDataObject obj, FileObject secondaryFile) {
        assert Arrays.binarySearch(OPTIONAL_EXT, secondaryFile.getExt().toLowerCase()) >= 0;
        FileEntry entry = new FileEntry(obj, secondaryFile);
        LOGGER.log(Level.FINEST, "Creating secondary entry for {0}", secondaryFile.getNameExt());
        return entry;
    }
    
    /**
     * Show a coverage in the default Geotools popup viewer.
     * Used for testing/comparison, should be disabled in production.
     */
    @ActionRegistration(displayName = "Geotools viewer", iconBase = GeoTiffDataObject.ICON_BASE)
    @ActionID(category = "View", id = "eu.agrosense.client.geotiff.GeotoolsShowAction")
    public static class GeotoolsShowAction extends AbstractAction {
        private final GeoTiffDataObject dob;
        
        public GeotoolsShowAction(GeoTiffDataObject dob) {
            this.dob = dob;
        }

        @Override
        public void actionPerformed(ActionEvent e) {
            dob.getCoverage().show(dob.getName());
        }
    }
    
}
