/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package eu.agrosense.client.geotiff;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.media.jai.PlanarImage;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import nl.cloudfarming.client.util.swing.MessagePanel;
import org.geotools.coverage.grid.GridCoverage2D;
import org.netbeans.core.spi.multiview.MultiViewElement;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.windows.TopComponent;

/**
 * Panel showing the source image for a GeoTIFF raster.
 * 
 * TODO: add pan/zoom
 * 
 * @author johan
 */
@MultiViewElement.Registration(
    displayName = "#LBL_Geotiff_IMAGE",
iconBase = GeoTiffDataObject.ICON_BASE,
mimeType = GeoTiffDataObject.MIME_TYPE,
persistenceType = TopComponent.PERSISTENCE_NEVER,
preferredID = "GeotiffVisual",
position = 1000)
@NbBundle.Messages("LBL_Geotiff_IMAGE=Image")
public class GeoTiffImageElement extends AbstractGeoTiffElement {
    private static final Logger LOGGER = Logger.getLogger("eu.agrosense.client.geotiff.GeoTiffImageElement");
    
    private final GridCoverage2D coverage;
    private PlanarImage planarImage;
    private BufferedImage image;
    private JScrollPane scrollPane;
    private final MessagePanel messagePanel = new MessagePanel(MessagePanel.LARGE_ICON);
    
    private boolean hasIndexOutOfBounds = false;// workaround until AGROSENSE-1494 is implemented.

    public GeoTiffImageElement(Lookup lkp) {
        super(lkp);
        coverage = obj.getCoverage();
        init();
    }
    
    private void init() {
        planarImage = (PlanarImage)coverage.getRenderedImage();
        
        LOGGER.log(Level.FINEST, "tiles: {0}x{1} (cols*rows)", new Object[]{planarImage.getNumXTiles(),planarImage.getNumYTiles()});
        LOGGER.log(Level.FINEST, "tile size: {0}x{1} (w*h)", new Object[]{planarImage.getTileWidth(),planarImage.getTileHeight()});        
        LOGGER.log(Level.FINEST, "color model: {0}", planarImage.getColorModel().toString());
        
        // DataBuffer: TYPE_BYTE = 0; TYPE_USHORT = 1; TYPE_SHORT = 2;
        // FIXME:
        // 16-bit grayscale returns type 2, but that can only use a 15-bit range... (see java.awt.image.ComponentColorModel.extractComponent())
        // looks like Short.MIN_VALUE is a no-data marker, add it to the colorMap/rasterSymbolizer.
        LOGGER.log(Level.FINEST, "transfer type: {0}", Integer.toString(planarImage.getColorModel().getTransferType()));
        
        JPanel imagePanel = new JPanel() {
            @Override
            protected void paintComponent(Graphics g) {
                super.paintComponent(g);
                
                if (hasIndexOutOfBounds) {return;}

                Rectangle rect = scrollPane.getViewport().getViewRect();
                image  = getImage(rect);
                
                try {
                    if (!imageFits(rect)) {
                        g.drawImage(image, rect.x, rect.y, this);
                    } else {
                        // center:
                        int x = (getWidth() - image.getWidth(this))/2;
                        int y = (getHeight() - image.getHeight(this))/2;
                        g.drawImage(image, x, y, this);
                    }
                } catch (ArrayIndexOutOfBoundsException ex) {
                    // workaround until AGROSENSE-1494 is implemented.
                    hasIndexOutOfBounds = true;
                    messagePanel.setError(ex.toString());
                }
            }
        };
        imagePanel.setPreferredSize(new Dimension(planarImage.getWidth(), planarImage.getHeight()));
        
        scrollPane = new JScrollPane(imagePanel);
        scrollPane.getViewport().setScrollMode(JViewport.BACKINGSTORE_SCROLL_MODE);

        setLayout(new BorderLayout());
        add(messagePanel, BorderLayout.NORTH);
        add(scrollPane, BorderLayout.CENTER);
        

        // BufferedImage: TYPE_BYTE_GRAY = 10; TYPE_USHORT_GRAY = 11; TYPE_CUSTOM = 0 means unrecognized
//        LOGGER.log(Level.FINEST, "image type: {0}", Integer.toString(image.getType()));
    }
        
    @Override
    public String getName() {
        return "GeoTiffVisualElement";
    }
    
    private BufferedImage getImage(Rectangle rect) {
        
        // scaling via a RenderableImage fails on an NPE (AbstractCoverage.java:710),
        // because it depends on the GridSampleDimension having categories.
//        PlanarImage scaledIfTooLarge = planarImage;
//        if (planarImage.getHeight() > scrollPane.getHeight()) {
//            RenderableImage renderable = coverage.getRenderableImage(
//                    coverage.getGridGeometry().axisDimensionX,
//                    coverage.getGridGeometry().axisDimensionY);
//            scaledIfTooLarge = (PlanarImage)renderable.createScaledRendering(0, h, null);
//        }

        return planarImage.getAsBufferedImage(imageFits(rect) ? null :  rect, null);
    }
    
    private boolean imageFits(Rectangle r) {
        return planarImage.getHeight() < r.getHeight() && planarImage.getWidth() < r.getWidth();
    }
}
