/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package eu.agrosense.client.geotiff;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Icon;
import nl.cloudfarming.client.imp.AbstractImportFolderNode;
import nl.cloudfarming.client.imp.AbstractImportProcessor;
import nl.cloudfarming.client.imp.ImportProcessor;
import nl.cloudfarming.client.imp.ImportSource;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.nodes.FilterNode;
import org.openide.nodes.Node;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;

/**
 *
 * @author johan
 */
@NbBundle.Messages("geotiff import format description=GeoTIFF file (*.tif)")
@ServiceProvider(service = ImportProcessor.class)
public class GeoTiffImportProcesssor extends AbstractImportProcessor {
    public static final String IMPORTED_GEOTIFF_FOLDER = "GeoTIFF";
    private static final Logger LOGGER = Logger.getLogger("eu.agrosense.client.geotiff.GeoTiffImportProcesssor");
    static final Icon ICON = ImageUtilities.loadImageIcon(GeoTiffDataObject.ICON_BASE, true);

    @Override
    public String[] getSupportedExtentions() {
        return new String[]{GeoTiffFileLoader.PRIMARY_EXT};
    }


    @Override
    public boolean isSupported(FileObject importFile) {
        return GeoTiffFileLoader.PRIMARY_EXT.equalsIgnoreCase(importFile.getExt());
    }

    @Override
    public String getFormatDescription() {
        return Bundle.geotiff_import_format_description();
    }

    @Override
    public Icon getIcon() {
        return ICON;
    }

    @Override
    public void doImport(FileObject importFile, Project project) throws IllegalArgumentException {
        assert isSupported(importFile) : "unsupported FileObject passed to doImport, check for support first";
        try {
            DataObject dataObject = DataObject.find(importFile);
            copySourceFiles(dataObject, getGeoTiffImportFolder(project));
        } catch (DataObjectNotFoundException ex) {
            LOGGER.log(Level.WARNING, "Unable to find dataobject for fileobject {0}", ex.getMessage());
        }
    }

    @Override
    public void process(FileObject importFile, Project project) throws IllegalArgumentException {
        // nothing to process.
    }

    @Override
    public <T> void process(FileObject importFile, Project project, List<T> toPopulate) throws IllegalArgumentException {
        // nothing to process.
    }

    @Override
    public Node getImportNode(Project project) {
        FileObject geotiffImportFolder = getGeoTiffImportFolder(project);
        DataFolder dataFolder = DataFolder.findFolder(geotiffImportFolder);
        assert dataFolder != null;
        return new ImportedGeoTiffsNode(project, dataFolder.getNodeDelegate());
    }
    
    private FileObject getGeoTiffImportFolder(Project project) {
        return getSubFolder(getImportRootFolder(project), IMPORTED_GEOTIFF_FOLDER);
    }

    @NbBundle.Messages("ImportedGeotiffsNode.name=GeoTIFF")
    private static final class ImportedGeoTiffsNode extends AbstractImportFolderNode {
        public ImportedGeoTiffsNode(Project project, Node original) {
            super(project, original, new ImportedGeoTiffsChildren(original));
            setName(Bundle.ImportedGeotiffsNode_name());
        }

        //TODO: create folder icons
//        @Override
//        public Image getIcon(int type) {
//            return ImageUtilities.loadImage("eu/agrosense/client/geotiff/folder_geotiff.png", true);
//        }
    }
    
    private static class ImportedGeoTiffsChildren extends FilterNode.Children {

        public ImportedGeoTiffsChildren(Node or) {
            super(or);
        }

        @Override
        protected Node[] createNodes(Node object) {
            List<Node> result = new ArrayList<>();

            for (Node node : super.createNodes(object)) {
                if (accept(node)) { result.add(node); }
            }
            return result.toArray(new Node[0]);
        }
        
        private boolean accept(Node node) {
            // hide secondary files:
            FileObject fo = node.getLookup().lookup(FileObject.class);
            if (fo == null) { return false; }
            return GeoTiffDataObject.MIME_TYPE.equals(fo.getMIMEType());
        }
    }

}
